/* $Id: dist.c,v 1.4 1999/10/03 09:00:45 galaxy-dev Exp $ */

/****h* GalaxyTools/Dist.c [2.0] *
*  NAME
*    Dist.c -- Compute Distances Between Planets (v1.0b)
*  COPYRIGHT
*    Maverick Software Development 1995
*    This software is public domain and can be freely redistributed as
*    long as it is in it's original state.
*  SYNOPSIS
*    dist -file        <planetfile> 
*         -lowerlimit  <lower limit>
*         -upperlimit  <upper limit>
*         -origin      <planet of origin>
*         [-sort  dist|name]
*  FUNCTION
*    Dist is a tool for finding the distances between planets in
*    the game Galaxy. Given a file with planet information,
*    it can find all planets that are more than <lower limit>
*    and less than <upper limit> away from specified planet
*    of origin.
*  INPUTS
*    planetfile      -- file with planet names and coordinates.
*                       Should have the following format:
*                       {name} {x-coordinate} { y-coordiante}
*                       You can generate this file with the 
*                       planetmap command.
*                         planetmap <galaxy report file>
*                         cat planets.* > planets.all
*
*    lower limit     -- planets that have a distance lower that
*                       this limit are not shown. 
*    upper limit     -- planets that have a distance greater that
*                       this limit are not shown.
*    planet_of_orgin -- the planet of origin.
*    sort flag       -- name, sort planets by name.
*                       dist, sort planets by distance.
*  RESULTS
*    A list with all names, coordinates, and distances.
*  EXAMPLE
*
*    Dist -file planets -lowerlimit 10 -upperlimit 50  -origin Zappa
*
*    Finds all planets in the range of [10,50> light years from
*    planet Zappa, and show their distance from planet Zappa.
*  AUTHOR
*    Frans Slothouber and Ken Weinert
*  CREATION DATE
*    06-May-1995
*  MODIFICATION HISTORY
*    06-May-1995  - version 1.0b
*    03-Oct-1999  - version 2.0
*                 Modifications by Ken:
*                 o Added better handling of options.
*                 Modifications by Frans:
*                 o Added a check to see whether the planet of
*                   origin is specified.
*
*  BUGS
*    Catch them in a jar and send them to fslothouber@acm.org
*  SEE ALSO
*    planetmap
**********
*/


#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "./getopt.h"

#define TRUE  1
#define FALSE 0

#define LINE_BUFFER_LENGTH 100
char            line_buffer[LINE_BUFFER_LENGTH];

typedef struct {
  float           dist;
  float           x;
  float           y;
  char            name[32];
} planet;

planet         *planetList = NULL;
int             nbrPlanetsAllocated = 0;
int             nbrPlanetsRead = 0;

int             load_planets(FILE *);
void            distances(planet *, planet *, float, float, char *);

int
nameComp(const void *item1, const void *item2)
{
  return strcmp(((planet *) item1)->name,
		((planet *) item2)->name);
}

int
distComp(const void *item1, const void *item2)
{
  if (((planet *) item1)->dist < ((planet *) item2)->dist) {
    return -1;
  }
  if (((planet *) item1)->dist == ((planet *) item2)->dist) {
    return 0;
  }
  if (((planet *) item1)->dist > ((planet *) item2)->dist) {
    return 1;
  }
}

int
main(int argc, char **argv)
{
  FILE           *planet_specs = NULL;
  planet         *the_planets = NULL;
  planet         *cur_planet;
  char            origin_planet[32];
  char            infilename[1024];

  int             return_code = 0;
  int             sortby = 1;	/* default is sort by name */
  int             optchar;
  int             option_index;
  int             err = 0;
  int             i;

  struct option   long_options[] =
  {
    {"file", required_argument, 0, 'f'},
    {"lowerlimit", required_argument, 0, 'l'},
    {"upperlimit", required_argument, 0, 'u'},
    {"origin", required_argument, 0, 'o'},
    {"sort", required_argument, 0, 's'},
    {0, 0, 0, 0}
  };

  double          low_range = -1.;
  double          up_range = -1.;

  infilename[0]    = '\0';
  origin_planet[0] = '\0';

  while ((optchar = getopt_long_only(argc, argv, "f:l:o:u:s:",
				  long_options, &option_index)) != EOF) {
    switch (optchar) {
    case 'f':			/* filename */
      strcpy(infilename, optarg);
      break;

    case 'l':			/* lowerlimit */
      low_range = atof(optarg);
      break;

    case 'u':			/* upper limit */
      up_range = atof(optarg);
      break;

    case 'o':			/* origin */
      if (isdigit(*optarg)) {
	sprintf(origin_planet, "%4d", atoi(optarg));
      } else {
	strcpy(origin_planet, optarg);
      }
      break;

    case 's':			/* sort by */
      if (strcmp(optarg, "dist") == 0) {
	sortby = 0;
      } else if (strcmp(optarg, "name") == 0) {
	sortby = 1;
      }
    }
  }

  if (infilename[0] == '\0') {
    fprintf(stderr, "No input file specified\n");
    err++;
  }
  if (origin_planet[0] == '\0') {
    fprintf(stderr, "No planet of origin specified\n");
    err++;
  }
  if (low_range < 0) {
    fprintf(stderr, "lower limit either not specified or less than 0\n");
    err++;
  }
  if (up_range < 0) {
    fprintf(stderr, "upper limit either not specified or less than 0\n");
    err++;
  }
  if (up_range < low_range) {
    double          tmp;

    tmp = low_range;
    low_range = up_range;
    up_range = tmp;
    fprintf(stderr, "lower/upper limits switched, I put them in the right order.\n");
  }
  if (err) {
    fprintf(stderr, "usage: dist -file InputFileName -lowerlimit LowLimit -upperlimit UpperLimit -origin planet_of_origin [-sort dist|name]\n");
    fprintf(stderr, "       You may use just the first letter of the option.\n");
    fprintf(stderr, "       The default sort is by name, ascending. The -sort dist option will sort by distance, ascending.\n");
    exit(EXIT_FAILURE);
  }
  if (planet_specs = fopen(infilename, "r")) {
    if (load_planets(planet_specs)) {
      distances(the_planets, the_planets, low_range, up_range, origin_planet);
    } else {
      printf("ERROR: Failed to make planet list.\n");
      return_code = 102;
    }
  } else {
    printf("ERROR: can't open file %s.\n", argv[1]);
    return_code = 101;
  }

  switch (sortby) {
  case 0:			/* distance */
    qsort((char *) planetList, nbrPlanetsRead, sizeof(planet), distComp);
    break;

  case 1:			/* name sort */
    qsort((char *) planetList, nbrPlanetsRead, sizeof(planet), nameComp);
    break;
  }

  for (i = 0; i < nbrPlanetsRead; i++) {
    if (planetList[i].dist > low_range && planetList[i].dist <= up_range) {
      printf("%-32.32s %7.2f %7.2f %7.2f\n", planetList[i].name,
	     planetList[i].x, planetList[i].y, planetList[i].dist);
    }
  }

  if (planet_specs) {
    fclose(planet_specs);
  }
  free(planetList);
  return return_code;
}


void 
distances(planet * from_planet,
	  planet * to_planet,
	  float low_range,
	  float up_range,
	  char *origin_planet)
{
  int             i, j;

  printf("Planet of Orgin %s ", origin_planet);
  for (i = 0; i < nbrPlanetsRead; i++) {
    if (strcmp(origin_planet, planetList[i].name) == 0) {
      printf(" (%7.2f %7.2f)", planetList[i].x, planetList[i].y);
      break;
    }
  }

  printf("\nPlanets in the range [%3.3f,%3.3f> from this planet.\n",
	 low_range, up_range);
  printf("%-32.32s     x       y   distance\n", "Name");
  for (j = 0; j < nbrPlanetsRead; j++) {
    float           t1, t2, distance;

    t1 = planetList[i].x - planetList[j].x;
    t2 = planetList[i].y - planetList[j].y;
    distance = sqrt(t1 * t1 + t2 * t2);
    planetList[j].dist = distance;
  }
}


int
load_planets(FILE * planet_specs)
{
  char            name[32];
  int             status = TRUE;

  fprintf(stderr, "Loading Planets\n");

  if (nbrPlanetsAllocated == 0) {
    nbrPlanetsAllocated = 100;
    planetList = (planet *) malloc(nbrPlanetsAllocated * sizeof(planet));
  }
  for (;
       !(feof(planet_specs)) && status;
      ) {
    if (nbrPlanetsRead == nbrPlanetsAllocated) {
      nbrPlanetsAllocated *= 2;
      planetList = (planet *) realloc(planetList, nbrPlanetsAllocated * sizeof(planet));
    }
    *line_buffer = '\n';
    fgets(line_buffer, LINE_BUFFER_LENGTH - 2, planet_specs);
    if (!isspace(*line_buffer)) {
      sscanf(line_buffer, "%s %f %f", name,
	   &planetList[nbrPlanetsRead].x, &planetList[nbrPlanetsRead].y);
      planetList[nbrPlanetsRead].dist = 0.;
      if (isdigit(*name)) {
	sprintf(planetList[nbrPlanetsRead].name, "%4d", atoi(name));
      } else {
	strcpy(planetList[nbrPlanetsRead].name, name);
      }
      nbrPlanetsRead++;
    }
  }
  return status;
}





